@@ -13,8 +13,8 @@
 package Messaging::Message;
 use strict;
 use warnings;
-our $VERSION  = "1.0";
-our $REVISION = sprintf("%d.%02d", q$Revision: 1.13 $ =~ /(\d+)\.(\d+)/);
+our $VERSION  = "1.2_1";
+our $REVISION = sprintf("%d.%02d", q$Revision: 1.18 $ =~ /(\d+)\.(\d+)/);
 
 #
 # export control
@@ -41,11 +41,19 @@ use Params::Validate qw(validate validate_pos :types);
 
 our(
     %_LoadedModule,		# hash of successfully loaded modules
+    %_CompressionModule,	# known compression modules
+    $_CompressionAlgos,		# known compression algorithms
     %_ValidateSpec,		# specifications for validate()
     %_ValidateType,		# types for validate()
     $_JSON,			# JSON object
 );
 
+%_CompressionModule = (
+    "lz4"    => "LZ4",
+    "snappy" => "Snappy",
+    "zlib"   => "Zlib",
+);
+$_CompressionAlgos = join("|", sort(keys(%_CompressionModule)));
 $_JSON = JSON->new();
 
 #
@@ -127,7 +135,7 @@ sub _eval ($&@) {
 sub _maybe_base64_encode ($) {
     my($object) = @_;
 
-    return unless $object->{body} =~ /[\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\xff]/;
+    return unless $object->{body} =~ /[^\t\n\r\x20-\x7e]/;
     # only if it contains more than printable ASCII characters (plus \t \n \r)
     _eval("Base64 encoding", sub {
 	$object->{body} = encode_base64($object->{body}, "");
@@ -139,32 +147,24 @@ sub _maybe_utf8_encode ($) {
     my($object) = @_;
     my($tmp);
 
-    return unless $object->{body} =~ /[\x80-\xff]/;
-    # only if it contains more than ASCII characters...
     _eval("UTF-8 encoding", sub {
 	$tmp = Encode::encode("UTF-8", $object->{body}, Encode::FB_CROAK|Encode::LEAVE_SRC);
     });
-    return if $object->{body} eq $tmp;
-    # ... and is worth encoding
+    return if $tmp eq $object->{body};
     $object->{body} = $tmp;
     $object->{encoding}{utf8}++;
-}
+			}
 
-sub _maybe_zlib_compress ($) {
-    my($object) = @_;
-    my($len, $tmp);
+sub _do_compress ($$) {
+    my($object, $algo) = @_;
+    my($compress, $tmp);
 
-    $len = length($object->{body});
-    return unless $len > 255;
-    # only if it is long enough...
-    _eval("Zlib compression", sub {
-	$tmp = Compress::Zlib::compress(\$object->{body});
+    $compress = \&{"Compress::$_CompressionModule{$algo}::compress"};
+    _eval("$_CompressionModule{$algo} compression", sub {
+	$tmp = $compress->(\$object->{body});
     });
-    # FIXME: for text body, we may loose space because of utf+base64... check final length?
-    return unless (length($tmp) / $len) < 0.9;
-    # ... and is worth compressing
     $object->{body} = $tmp;
-    $object->{encoding}{zlib}++;
+    $object->{encoding}{$algo}++;
 }
 
 #+++############################################################################
@@ -310,45 +310,66 @@ sub size : method {
 #
 
 $_ValidateSpec{jsonify} = {
-    compression => { type => SCALAR, regex => qr/^(zlib)?$/, optional => 1 },
+    compression => { type => SCALAR, regex => qr/^($_CompressionAlgos)?!?$/o, optional => 1 },
 };
 
 sub jsonify : method {
-    my($self, %option, $compression, %object);
+    my($self, %option, %object, $len, $algo, $force);
 
     $self = shift(@_);
     %option = validate(@_, $_ValidateSpec{jsonify}) if @_;
-    $compression = $option{compression} || "";
+    ($algo, $force) = ($1, $2)
+	if $option{compression} and $option{compression} =~ /^(\w+)(!?)$/;
     # check compression availability
-    _require("Compress::Zlib") if $compression eq "zlib";
+    _require("Compress::$_CompressionModule{$algo}") if $algo;
     # build the JSON object
     $object{text} = JSON::true if $self->{text};
     $object{header} = $self->{header} if keys(%{ $self->{header} });
-    return(\%object) unless length(${ $self->{body_ref} });
+    $len = length(${ $self->{body_ref} });
+    return(\%object) unless $len;
     $object{body} = ${ $self->{body_ref} };
     # handle non-empty body
     if ($self->{text}) {
 	# text body
-	if ($compression) {
+	if ($algo and $force) {
+	    # always compress
 	    _maybe_utf8_encode(\%object);
-	    if ($compression eq "zlib") {
-		_maybe_zlib_compress(\%object);
-	    }
-	    if ($object{encoding} and $object{encoding}{zlib}) {
-		# we did compress
-		_maybe_base64_encode(\%object);
-	    } else {
-		# in fact, we did not compress
+	    _do_compress(\%object, $algo);
+	    _maybe_base64_encode(\%object);
+	} elsif ($algo and $len > 255) {
+	    # maybe compress
+	    _maybe_utf8_encode(\%object);
+	    _do_compress(\%object, $algo);
+	    _maybe_base64_encode(\%object);
+	    if (length($object{body}) >= $len) {
+		# not worth it
 		$object{body} = ${ $self->{body_ref} };
 		delete($object{encoding});
 	    }
+	} else {
+	    # do not compress
 	}
     } else {
 	# binary body
-	if ($compression eq "zlib") {
-	    _maybe_zlib_compress(\%object);
+	if ($algo and $force) {
+	    # always compress
+	    _do_compress(\%object, $algo);
+	    _maybe_base64_encode(\%object);
+	} elsif ($algo and $len > 255) {
+	    # maybe compress
+	    $len *= 4/3 if $object{body} =~ /[^\t\n\r\x20-\x7e]/;
+	    _do_compress(\%object, $algo);
+	    _maybe_base64_encode(\%object);
+	    if (length($object{body}) >= $len) {
+		# not worth it
+		$object{body} = ${ $self->{body_ref} };
+		delete($object{encoding});
+		_maybe_base64_encode(\%object);
+	    }
+	} else {
+	    # do not compress
+	    _maybe_base64_encode(\%object);
 	}
-	_maybe_base64_encode(\%object);
     }
     # set the encoding string
     $object{encoding} = join("+", sort(keys(%{ $object{encoding} })))
@@ -369,7 +390,7 @@ $_ValidateSpec{dejsonify} = {
 };
 
 sub dejsonify : method {
-    my($class, $object, $encoding, $self, $tmp, $len);
+    my($class, $object, $encoding, $self, $tmp, $len, $uncompress);
 
     $class = shift(@_);
     validate_pos(@_, { type => HASHREF })
@@ -378,8 +399,8 @@ sub dejsonify : method {
     $object = $_[0];
     $encoding = $object->{encoding} || "";
     _fatal("invalid encoding: %s", $encoding)
-	unless $encoding eq "" or "${encoding}+" =~ /^((base64|utf8|zlib)\+)+$/;
-    _require("Compress::Zlib") if $encoding =~ /zlib/;
+	unless $encoding eq "" or "${encoding}+" =~ /^((base64|utf8|$_CompressionAlgos)\+)+$/o;
+    _require("Compress::$_CompressionModule{$1}") if $encoding =~ /($_CompressionAlgos)/o;
     # construct the message
     $self = $class->new();
     $self->{text} = 1 if $object->{text};
@@ -398,17 +419,19 @@ sub dejsonify : method {
 	    });
 	    _fatal("invalid Base64 data: %s", $object->{body}) unless $len == length($tmp);
 	}
-	if ($encoding =~ /zlib/) {
-	    # body has been Zlib compressed
-	    _eval("Zlib decompression", sub {
-		$tmp = Compress::Zlib::uncompress(\$tmp);
+	if ($encoding =~ /($_CompressionAlgos)/o) {
+	    # body has been compressed
+	    $uncompress = \&{"Compress::$_CompressionModule{$1}::uncompress"};
+	    _eval("$_CompressionModule{$1} decompression", sub {
+		$tmp = $uncompress->(\$tmp);
 	    });
+	    _fatal("invalid $_CompressionModule{$1} compressed data!") unless defined($tmp);
 	}
 	if ($encoding =~ /utf8/) {
 	    # body has been UTF-8 encoded
 	    _eval("UTF-8 decoding", sub {
-		$tmp = Encode::decode("UTF-8", $tmp, Encode::FB_CROAK|Encode::LEAVE_SRC);
-	    }) if $tmp =~ /[\x80-\xff]/;
+		$tmp = Encode::decode("UTF-8", $tmp, Encode::FB_CROAK);
+	    });
 	}
 	$self->{body_ref} = \$tmp;
     }
@@ -491,9 +514,8 @@ sub serialize : method {
 
     $self = shift(@_);
     $tmp = $self->stringify_ref(@_);
-    return($$tmp) unless $$tmp =~ /[\x80-\xff]/;
     _eval("UTF-8 encoding", sub {
-	$tmp = Encode::encode("UTF-8", $$tmp, Encode::FB_CROAK|Encode::LEAVE_SRC);
+	$tmp = Encode::encode("UTF-8", $$tmp, Encode::FB_CROAK);
     });
     return($tmp);
 }
@@ -503,9 +525,8 @@ sub serialize_ref : method {
 
     $self = shift(@_);
     $tmp = $self->stringify_ref(@_);
-    return($tmp) unless $$tmp =~ /[\x80-\xff]/;
     _eval("UTF-8 encoding", sub {
-	$tmp = Encode::encode("UTF-8", $$tmp, Encode::FB_CROAK|Encode::LEAVE_SRC);
+	$tmp = Encode::encode("UTF-8", $$tmp, Encode::FB_CROAK);
     });
     return(\$tmp);
 }
@@ -520,7 +541,7 @@ sub deserialize : method {
     $class = shift(@_);
     validate_pos(@_, { type => SCALAR })
 	unless @_ == 1 and defined($_[0]) and ref($_[0]) eq "";
-    return($class->destringify($_[0])) unless $_[0] =~ /[\x80-\xff]/;
+    return($class->destringify($_[0])) unless $_[0] =~ /[^[:ascii:]]/;
     _eval("UTF-8 decoding", sub {
 	$tmp = Encode::decode("UTF-8", $_[0], Encode::FB_CROAK|Encode::LEAVE_SRC);
     }, @_);
@@ -533,7 +554,7 @@ sub deserialize_ref : method {
     $class = shift(@_);
     validate_pos(@_, { type => SCALARREF })
 	unless @_ == 1 and defined($_[0]) and ref($_[0]) eq "SCALAR";
-    return($class->destringify_ref($_[0])) unless ${ $_[0] } =~ /[\x80-\xff]/;
+    return($class->destringify_ref($_[0])) unless ${ $_[0] } =~ /[^[:ascii:]]/;
     _eval("UTF-8 decoding", sub {
 	$tmp = Encode::decode("UTF-8", ${ $_[0] }, Encode::FB_CROAK|Encode::LEAVE_SRC);
     }, @_);
@@ -657,17 +678,17 @@ that can be:
 
 =over
 
-=item base64
+=item C<base64>
 
 Base64 encoding (for binary body or compressed body)
 
-=item utf8
+=item C<utf8>
 
 UTF-8 encoding (only needed for a compressed text body)
 
-=item zlib
+=item C<snappy> or C<lz4> or C<zlib>
 
-Zlib compression
+Snappy or LZ4 or Zlib compression (only one can be specified)
 
 =back
 
@@ -846,12 +867,18 @@ of the header and body
 
 =back
 
-The jsonify(), stringify() and serialize() methods can be given
-options. Currently, the only supported option is C<compression> and
-the only supported compression is C<zlib> (when available). Here is
-for instance how to serialize a message, with compression:
+The jsonify(), stringify() and serialize() methods can be given options.
 
-  $bytes = $msg->serialize(compression => "zlib");
+Currently, the only supported option is C<compression> and it can
+contain either an algorithm name like C<zlib> (meaning: use this
+algorithm only of the compressed body is indeed smaller) or an
+algorithm name followed by an exclamation mark to always force
+compression.
+
+Here is for instance how to serialize a message, with forced
+compression:
+
+  $bytes = $msg->serialize(compression => "zlib!");
 
 In addition, in order to avoid string copies, the following methods
 are also available:
@@ -881,6 +908,8 @@ strings. Here is an example:
 
 =head1 SEE ALSO
 
+L<Compress::Snappy>,
+L<Compress::LZ4>,
 L<Compress::Zlib>,
 L<Encode>,
 L<JSON>.
